from typing import Callable
from typing import List
from typing import Optional

from smart_injector.config.backend import Bindings
from smart_injector.config.backend import ConfigBackend
from smart_injector.config.backend import ConfigEntry
from smart_injector.config.backend import Dependencies
from smart_injector.config.backend import FactoryArgs
from smart_injector.config.backend import Instances
from smart_injector.config.backend import Lifetimes
from smart_injector.config.user import Config
from smart_injector.container.container import StaticContainer
from smart_injector.lifetime import Lifetime
from smart_injector.resolver.handlers import AbstractTypeHandler
from smart_injector.resolver.handlers import BindingHandler
from smart_injector.resolver.handlers import BuiltinsTypeHandler
from smart_injector.resolver.handlers import InstanceFactory
from smart_injector.resolver.handlers import InstanceHandler
from smart_injector.resolver.handlers import NewInstanceHandler
from smart_injector.resolver.handlers import SingletonBaseTypeHandler
from smart_injector.resolver.handlers import SingletonEffectiveHandler
from smart_injector.resolver.resolver import Resolver


def create_container(
    configure: Optional[Callable[[Config], None]] = None,
    default_lifetime=Lifetime.TRANSIENT,
    dependencies: Optional[List[object]] = None,
) -> StaticContainer:
    """
    Use this function to create a DI container.

    :param configure:
    :param default_lifetime:
    :param dependencies:
    :return:
    """
    if configure is None:
        configure = _default_config
    if dependencies is None:
        dependencies = []
    backend = _create_backend(default_lifetime)
    resolver = _create_resolver(backend)
    container = StaticContainer(resolver=resolver)
    configure(Config(backend=backend))
    _resolve_dependencies(backend, dependencies)
    return container


def _default_config(config: Config):
    pass


def _create_backend(default_lifetime: Lifetime) -> ConfigBackend:
    lifetimes = Lifetimes(default_lifetime)
    instances = Instances()
    bindings = Bindings()
    factory_args = FactoryArgs()
    _dependencies = Dependencies()
    return ConfigBackend(bindings, lifetimes, instances, factory_args, _dependencies)


def _create_resolver(backend: ConfigBackend):
    resolver = Resolver()
    instance_factory = InstanceFactory(resolver, backend.factory_args)
    resolver.add_type_handler(InstanceHandler(backend.instances))
    resolver.add_type_handler(BindingHandler(resolver, backend.bindings))
    resolver.add_type_handler(
        SingletonBaseTypeHandler(backend.lifetimes, backend.instances, instance_factory)
    )
    resolver.add_type_handler(
        SingletonEffectiveHandler(
            backend.lifetimes, backend.instances, instance_factory
        )
    )
    resolver.add_type_handler(AbstractTypeHandler())
    resolver.add_type_handler(
        BuiltinsTypeHandler(my_builtins=[int, float, str, bytearray, bytes])
    )
    resolver.add_type_handler(NewInstanceHandler(instance_factory))
    return resolver


def _resolve_dependencies(backend: ConfigBackend, dependencies: List[object]):
    for dependency in dependencies:
        dependent_type = type(dependency)
        if dependent_type not in backend.dependencies.get_dependencies():
            raise TypeError(
                "no dependency was declared for type {dependent}".format(
                    dependent=type(dependency)
                )
            )
        backend.instances.set_instance(ConfigEntry(dependent_type), dependency)
        backend.dependencies.remove_dependency(dependent_type)

    if backend.dependencies.get_dependencies():
        raise TypeError(
            "no dependency were declared for the types  {dependency}".format(
                dependency=backend.dependencies.get_dependencies()
            )
        )



import inspect
from abc import abstractmethod
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Type

from smart_injector.config.backend import Bindings
from smart_injector.config.backend import ConfigEntry
from smart_injector.config.backend import ConfigVisibility
from smart_injector.config.backend import FactoryArgs
from smart_injector.config.backend import Instances
from smart_injector.config.backend import Lifetimes
from smart_injector.container.container import S
from smart_injector.container.container import T
from smart_injector.resolver.resolver import Resolver
from smart_injector.types import Handler
from smart_injector.types import ResolveRequest


def dependencies(a_type: Callable[..., T]) -> Dict[str, Type[Any]]:
    return {
        parameter.name: parameter.annotation
        for parameter in inspect.signature(a_type).parameters.values()
    }


class InstanceHandler(Handler):
    def __init__(self, instances: Instances):
        self._instances = instances

    def can_handle_type(self, request: ResolveRequest) -> bool:
        return (
            True
            if self._instances.has_instance(request.local_config_entry())
            else False
        )

    def handle(self, request: ResolveRequest) -> T:
        return self._instances.get_instance(request.local_config_entry())


class BindingHandler(Handler):
    def __init__(self, resolver: Resolver, bindings: Bindings):
        self._resolver = resolver
        self._bindings = bindings

    def can_handle_type(self, request: ResolveRequest) -> bool:
        return (
            True
            if request.real_type
            is not self._bindings.get_binding(request.local_config_entry())
            else False
        )

    def handle(self, request: ResolveRequest) -> S:
        return self._resolver.get_new_instance(
            request.new_request_with_same_origin(
                self._bindings.get_binding(request.local_config_entry())
            )
        )


class AbstractTypeHandler(Handler):
    def can_handle_type(self, request: ResolveRequest) -> bool:
        return True if inspect.isabstract(request.real_type) else False

    def handle(self, request: ResolveRequest) -> T:
        raise TypeError("No binding for abstract base {0}".format(request.real_type))


class BuiltinsTypeHandler(Handler):
    def __init__(self, my_builtins: List[Type[Any]]):
        self._my_builtins = my_builtins

    def can_handle_type(self, request: ResolveRequest) -> bool:
        return True if request.real_type in self._my_builtins else False

    def handle(self, request: ResolveRequest) -> T:
        return request.real_type()


class InstanceFactory:
    def __init__(self, resolver: Resolver, args: FactoryArgs):
        self._resolver = resolver
        self._args = args

    def create(self, context: ResolveRequest) -> T:
        return context.real_type(
            **self._dependency_instances(context), **self._factory_args(context)
        )

    def _factory_args(self, context: ResolveRequest) -> Dict[str, Any]:
        return {
            name: value.get(self._resolver)
            for name, value in self._args.get_factory_args(
                context.local_config_entry()
            ).items()
        }

    def _dependency_instances(self, context: ResolveRequest) -> Dict[str, Any]:
        return {
            name: self._resolver.get_new_instance(
                context.get_new_dependency_context(dependent)
            )
            for name, dependent in dependencies(context.real_type).items()
            if name not in self._args.get_factory_args(context.local_config_entry())
        }


class NewInstanceHandler(Handler):
    def __init__(self, factory: InstanceFactory):
        self._factory = factory

    def can_handle_type(self, request: ResolveRequest) -> bool:
        return True

    def handle(self, request: ResolveRequest) -> T:
        return self._factory.create(request)


class SingletonHandler(Handler):
    def __init__(
        self,
        lifetimes: Lifetimes,
        instances: Instances,
        instance_factory: InstanceFactory,
    ):
        self._lifetimes = lifetimes
        self._instances = instances
        self._instance_factory = instance_factory

    def can_handle_type(self, request: ResolveRequest) -> bool:
        return True if self._is_singleton(request) else False

    def handle(self, request: ResolveRequest) -> T:
        if self._singleton_not_created(request):
            self._create_singleton(request)
        return self._get_singleton(request)

    def _is_singleton(self, context: ResolveRequest):
        return self._lifetimes.is_singleton(self._local_config_entry(context))

    def _singleton_not_created(self, context: ResolveRequest) -> bool:
        return not self._instances.has_instance(context.local_config_entry())

    def _create_singleton(self, context: ResolveRequest):
        self._instances.set_instance(
            self._instance_context(context), self._instance_factory.create(context)
        )

    def _instance_context(self, context: ResolveRequest) -> ConfigEntry:
        if (
            self._lifetimes.visibility(context.local_config_entry())
            is ConfigVisibility.LOCAL
        ):
            return self._local_config_entry(context)
        else:
            return self._global_config_entry(context)

    def _get_singleton(self, context: ResolveRequest) -> T:
        return self._instances.get_instance(self._local_config_entry(context))

    @abstractmethod
    def _local_config_entry(self, context: ResolveRequest) -> ConfigEntry:
        pass

    @abstractmethod
    def _global_config_entry(self, context: ResolveRequest) -> ConfigEntry:
        pass


class SingletonBaseTypeHandler(SingletonHandler):
    def _local_config_entry(self, context: ResolveRequest) -> ConfigEntry:
        return context.local_base_config_entry()

    def _global_config_entry(self, context: ResolveRequest) -> ConfigEntry:
        return context.global_base_config_entry()


class SingletonEffectiveHandler(SingletonHandler):
    def _local_config_entry(self, context: ResolveRequest) -> ConfigEntry:
        return context.local_config_entry()

    def _global_config_entry(self, context: ResolveRequest) -> ConfigEntry:
        return context.global_config_entry()



import inspect
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import cast

from smart_injector.config.backend import ConfigBackend
from smart_injector.config.backend import FactoryArg
from smart_injector.config.backend import ValueArg
from smart_injector.lifetime import Lifetime
from smart_injector.types import ConfigEntry
from smart_injector.utility import get_return_type

T = TypeVar("T")
S = TypeVar("S")

Where = Optional[Type[T]]


def is_base_class(base: Callable[..., T], subclass: Callable[..., S]) -> bool:
    if base in inspect.getmro(cast(type, subclass)):
        return True
    return False


def ensure_subclass(a_type: Callable[..., T], to_type: Optional[Callable[..., S]]):
    if to_type is None or not is_base_class(a_type, to_type):
        raise TypeError(
            "{to_type} must be a subclass of {a_type}".format(
                to_type=to_type, a_type=a_type
            )
        )


def ensure_binding(a_type: Callable[..., T], to_type: Callable[..., S]):
    if inspect.isclass(to_type):
        ensure_subclass(a_type, to_type)
    elif callable(a_type):
        if get_return_type(to_type) is None:
            raise TypeError(
                "type annotation for callable {to_type} is None or is missing".format(
                    to_type=to_type
                )
            )
        ensure_subclass(a_type, get_return_type(to_type))
    else:
        raise TypeError("{to_type} must be callable".format(to_type=to_type))


class Config:
    """Used by the user to configure DI container injection behaviour"""

    def __init__(self, backend: ConfigBackend):
        """
        Users should not directly create a config by invoking this contructor. Instead use
        :py:meth:`smart_injector.create_container`

        """
        self._backend = backend

    def bind(
        self, a_type: Callable[..., T], to_type: Callable[..., S], where: Where = None
    ):
        """Specify a binding. Whenever an object of type a_type is required, then an object of type to_type will be provided.
        For example you can configure, which concrete class shall be used for an abstract base class

        :param a_type: will be replaced by to_type
        :param to_type: will be used when an object of type a_type is required. to_type must be a subclass of a_type
        :param where:
        :param kwargs:

        :return:
        """
        ensure_binding(a_type, to_type)
        self._backend.bindings.set_binding(ConfigEntry(a_type, where), to_type)

    def lifetime(
        self, a_type: Callable[..., T], lifetime: Lifetime, where: Where = None
    ):
        """
        Specify the lifetime for an object of type `T`. See :py:meth:`smart_injector.Lifetime`

        :param a_type:
        :param lifetime:
        :param where:
        :return: None

        """
        self._backend.lifetimes.set_lifetime(ConfigEntry(a_type, where), lifetime)

    def instance(self, a_type: Callable[..., T], instance: T, where: Where = None):
        """
        set an instance of type `T` which is returned whenever an object of type `T` is requested

        :param a_type:
        :param instance:
        :param where:
        :return:

        """
        if not isinstance(instance, cast(Type[T], a_type)):
            raise TypeError(
                "{instance} is not an instance of type {a_type}".format(
                    instance=instance, a_type=a_type
                )
            )
        self._backend.instances.set_instance(ConfigEntry(a_type, where), instance)

    def dependency(self, a_type: Callable[..., T]):
        """
        declare that `T` is a dependency for the container. When creating the container with :py:meth:`smart_injector.create_container`
        an instance must be provided for every dependency which was declared.

        :param a_type:
        :return:

        """
        self._backend.dependencies.add_dependency(a_type)

    def arguments(self, a_type: Callable[..., T], where: Where = None, **kwargs: Any):
        """
        When creating an object of type `T`, the provided arguments will be inserted in `T`'s constructor (if it is a class) or
        a_type will be called with this arguments if it  is a function.

        :param a_type:
        :param where:
        :param kwargs:
        :return:

        .. Note:: Only keyword arguments are supported
        """
        ensure_arguments(a_type, kwargs)
        self._backend.factory_args.set_factory_args(
            ConfigEntry(a_type, where),
            {name: ValueArg(value) for name, value in kwargs.items()},
        )

    def arg_factory(
        self, a_type: Callable[..., T], where: Where = None, **kwargs: Callable[..., S]
    ):
        """
        In difference to :py:meth:`.arguments`:
        Instead of providing a value for `parameter` directly, `factory` is called
        to get the value for the parameter.

        :param a_type:
        :param where:
        :param kwargs:
        :return:
        """
        for parameter, factory in kwargs.items():
            ensure_parameter(a_type, parameter)
            self._backend.factory_args.set_factory_args(
                ConfigEntry(a_type, where), {parameter: FactoryArg(factory)}
            )


def ensure_arguments(a_type: Callable[..., T], kwargs: Dict[str, Any]):
    for parameter in kwargs:
        ensure_parameter(a_type, parameter)


def ensure_parameter(a_type: Callable[..., T], parameter: str):
    if not has_parameter(a_type, parameter):
        raise TypeError(
            "parameter {parameter} not expected for type {a_type}".format(
                parameter=parameter, a_type=a_type
            )
        )


def has_parameter(a_type: Callable[..., T], parameter: str) -> bool:
    expected_parameters = inspect.signature(a_type).parameters.keys()
    return True if parameter in expected_parameters else False



from typing import Callable
from typing import TypeVar

from smart_injector.resolver.resolver import Resolver

T = TypeVar("T")
S = TypeVar("S")


class StaticContainer:
    """DI Container. Used by the user to get instances of types.

    To get your own container. Create a new class inherited from this class and override configure method
    """

    def __init__(self, resolver: Resolver):
        """You should not create an instance of your DI container own your own. Use the factory function create_container
        instead"""
        self.__resolver = resolver

    def get(self, a_type: Callable[..., T]) -> T:
        """
        Get an instance of type `T`

        :param a_type: either a class `T` or a function returning a `T`
        :return: an instance of `T`
        """
        return self.__resolver.get_instance(a_type)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.ifconfig',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]
doctest_global_setup = """
from smart_injector import create_container, StaticContainer, Lifetime, Config
"""
source_suffix = '.rst'
master_doc = 'index'
project = 'Smart Injector'
year = '2019'
author = 'Hendrik Levering'
copyright = '{0}, {1}'.format(year, author)
version = release = '0.0.6'

pygments_style = 'trac'
templates_path = ['.']
extlinks = {
    'issue': ('https://github.com/hlevering/python-smart-injector/issues/%s', '#'),
    'pr': ('https://github.com/hlevering/python-smart-injector/pull/%s', 'PR #'),
}
import sphinx_rtd_theme
html_theme = "sphinx_rtd_theme"
html_theme_path = ["_themes", ]
#html_theme_options = {
#    'githuburl': 'https://github.com/hlevering/python-smart-injector/'
#}

html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = False
html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



import inspect
from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from enum import Enum
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import TypeVar
from typing import cast

from smart_injector.lifetime import Lifetime
from smart_injector.resolver.resolver import Resolver
from smart_injector.types import ConfigEntry


class ConfigVisibility(Enum):
    LOCAL = 0
    GLOBAL = 1


T = TypeVar("T")
S = TypeVar("S")
U = TypeVar("U")


class ContextConfig(Generic[U]):
    def __init__(self, default_factory: Callable[[ConfigEntry], U]):
        self._default_factory = default_factory
        self._default = {}  # type: Dict[Callable[..., T], U]
        self._with_context = cast(
            Dict[Callable[..., T], Dict[Callable[..., S], U]], defaultdict(dict)
        )

    def get_visibility(self, what: ConfigEntry) -> ConfigVisibility:
        if self.is_defined_locally(what):
            return ConfigVisibility.LOCAL
        else:
            return ConfigVisibility.GLOBAL

    def set(self, item: ConfigEntry, to_type: U):
        if item.where is None:
            self._default[item.a_type] = to_type
        else:
            self._with_context[item.where][item.a_type] = to_type

    def is_defined_locally(self, item: ConfigEntry) -> bool:
        if item.where is not None and item.a_type in self._with_context[item.where]:
            return True
        else:
            return False

    def get(self, item: ConfigEntry) -> U:
        if self.is_defined_locally(item):
            return self._with_context[cast(Callable[..., T], item.where)].get(
                item.a_type, self._default_factory(item)
            )
        return self._default.get(item.a_type, self._default_factory(item))

    def delete(self, item: ConfigEntry):
        if item.where is None:
            self._default.pop(item.a_type, cast(U, None))
        else:
            self._with_context[item.where].pop(item.a_type, cast(U, None))


class ArgProxy(ABC):
    @abstractmethod
    def get(self, resolver: Resolver):
        pass


class ValueArg(ArgProxy):
    def __init__(self, value: T):
        self.value = value

    def get(self, resolver: Resolver) -> T:
        return self.value


class Inspector:
    def __init__(self, object: Any):
        self._object = object

    @property
    def class_name(self) -> str:
        return ".".join(self._object.__qualname__.split(".")[:-1])

    @property
    def method_name(self) -> str:
        return self._object.__qualname__.split(".")[-1]

    @property
    def klass_type(self) -> Any:
        return self._object.__globals__[self.class_name]

    @property
    def parameters(self) -> List[Any]:
        return list(inspect.signature(self._object).parameters)


def method_of_not_created_class(method: Callable[..., T]) -> bool:
    inspector = Inspector(method)
    if inspect.isfunction(method):
        try:
            klass = inspector.klass_type
        except KeyError:
            return False
        if inspect.isclass(klass):
            parameters = inspector.parameters
            if parameters and parameters[0] == "self":
                return True
    return False


def create_class_and_call_method(method: Callable[..., T], resolver: Resolver) -> T:
    class_instance = resolver.get_instance(Inspector(method).klass_type)
    return resolver.get_instance(getattr(class_instance, Inspector(method).method_name))


class FactoryArg(ArgProxy):
    def __init__(self, factory: Callable[..., T]):
        self.factory = factory

    def get(self, resolver: Resolver) -> T:
        if method_of_not_created_class(self.factory):
            return create_class_and_call_method(self.factory, resolver)
        return resolver.get_instance(self.factory)


class FactoryArgs:
    def __init__(self):
        self._config = ContextConfig[Dict[str, ArgProxy]](lambda x: {})

    def set_factory_args(self, what: ConfigEntry, kwargs: Dict[str, ArgProxy]):
        args = self.get_factory_args(what)
        args.update(kwargs)
        self._config.set(what, args)

    def get_factory_args(self, what: ConfigEntry) -> Dict[str, ArgProxy]:
        return self._config.get(what)


class Instances:
    def __init__(self):
        self._config = ContextConfig[Optional[object]](lambda x: None)

    def set_instance(self, what: ConfigEntry, instance: T):
        self._config.set(what, instance)

    def has_instance(self, what: ConfigEntry) -> bool:
        return False if self._config.get(what) is None else True

    def get_instance(self, what: ConfigEntry) -> T:
        return cast(T, self._config.get(what))


class Bindings:
    def __init__(self):
        self._config = ContextConfig[Callable[..., T]](lambda x: x.a_type)

    def set_binding(self, what: ConfigEntry, to_type: Callable[..., S]):
        self._config.set(what, to_type)

    def get_binding(self, which: ConfigEntry) -> Callable[..., S]:
        return self._config.get(which)


class Lifetimes:
    def __init__(self, default_lifetime: Lifetime):
        self._config = ContextConfig[Callable[..., T]](
            lambda x: default_lifetime  # type: ignore
        )

    def is_singleton(self, what: ConfigEntry) -> bool:
        return (
            True
            if cast(Lifetime, self._config.get(what)) is Lifetime.SINGLETON
            else False
        )

    def set_lifetime(self, what: ConfigEntry, lifetime: Lifetime):
        self._remove_lifetime_setting(what)
        if lifetime is not Lifetime._INTERNAL_DEFAULT:
            self._config.set(what, lifetime)  # type: ignore

    def _remove_lifetime_setting(self, what: ConfigEntry):
        self._config.delete(what)

    def visibility(self, what: ConfigEntry) -> ConfigVisibility:
        return self._config.get_visibility(what)


class Dependencies:
    def __init__(self):
        self._dependencies = {}  # type: Dict[Callable[..., T], None]

    def add_dependency(self, a_type: Callable[..., T]):
        if a_type not in self._dependencies:
            self._dependencies[a_type] = None

    def remove_dependency(self, a_type: Callable[..., T]):
        self._dependencies.pop(a_type, None)

    def get_dependencies(self) -> Dict[Callable[..., T], None]:
        return cast(Dict[Callable[..., T], None], self._dependencies)


class ConfigBackend:
    """Simple container for all configuration classes"""

    def __init__(
        self,
        bindings: Bindings,
        lifetimes: Lifetimes,
        instances: Instances,
        factory_args: FactoryArgs,
        dependencies: Dependencies,
    ):
        self.bindings = bindings
        self.lifetimes = lifetimes
        self.instances = instances
        self.factory_args = factory_args
        self.dependencies = dependencies



from abc import ABC
from abc import abstractmethod
from typing import Callable
from typing import Optional
from typing import TypeVar

T = TypeVar("T")


class ConfigEntry:
    def __init__(
        self, a_type: Callable[..., T], where: Optional[Callable[..., T]] = None
    ):
        self.a_type = a_type
        self.where = where


class ResolveRequest:
    def __init__(
        self,
        real_type: Callable[..., T],
        base_type: Callable[..., T],
        where: Optional[Callable[..., T]],
    ):
        self.real_type = real_type
        self.base_type = base_type
        self.where = where

    def new_request_with_same_origin(
        self, a_type: Callable[..., T]
    ) -> "ResolveRequest":
        return ResolveRequest(a_type, self.base_type, where=self.where)

    def get_new_dependency_context(self, a_type: Callable[..., T]) -> "ResolveRequest":
        return ResolveRequest(a_type, a_type, where=self.base_type)

    def local_config_entry(self) -> ConfigEntry:
        return ConfigEntry(self.real_type, self.where)

    def global_config_entry(self) -> ConfigEntry:
        return ConfigEntry(self.real_type, where=None)

    def local_base_config_entry(self) -> ConfigEntry:
        return ConfigEntry(self.base_type, self.where)

    def global_base_config_entry(self) -> ConfigEntry:
        return ConfigEntry(self.base_type, where=None)


class Handler(ABC):
    @abstractmethod
    def can_handle_type(self, request: ResolveRequest) -> bool:
        pass

    @abstractmethod
    def handle(self, request: ResolveRequest) -> T:
        pass



from smart_injector.config.backend import ConfigEntry
from smart_injector.config.backend import ContextConfig


class A:
    pass


class B:
    pass


class C:
    pass


class D:
    pass


def test_context_register():
    register = ContextConfig(lambda x: x.a_type)
    register.set(ConfigEntry(A, B), C)
    register.set(ConfigEntry(A, None), D)
    t1 = register.get(ConfigEntry(A, B))
    t2 = register.get(ConfigEntry(A, None))
    assert t1 is C
    assert t2 is D
    register.delete(ConfigEntry(A, B))
    t3 = register.get(ConfigEntry(A, B))
    assert t3 is D
    register.delete(ConfigEntry(A, None))
    t4 = register.get(ConfigEntry(A, B))
    assert t4 is A



from abc import ABC
from abc import abstractmethod

import pytest  # type: ignore

from smart_injector import Lifetime
from smart_injector import StaticContainer
from smart_injector.config.backend import Bindings
from smart_injector.config.backend import ConfigBackend
from smart_injector.config.backend import Dependencies
from smart_injector.config.backend import FactoryArgs
from smart_injector.config.backend import Instances
from smart_injector.config.backend import Lifetimes
from smart_injector.config.backend import method_of_not_created_class
from smart_injector.config.user import Config
from smart_injector.container.factory import create_container
from smart_injector.resolver.resolver import Resolver


class A:
    pass


def test_container_creates_simple_type():
    container = create_container()
    my_int = container.get(A)
    assert isinstance(my_int, A)


@pytest.fixture
def container() -> StaticContainer:
    return create_container()


class B:
    pass


class AB:
    def __init__(self, a: A, b: B):
        self.a = a
        self.b = b


def test_container_resolves_concrete_types(container: StaticContainer) -> None:
    ab = container.get(AB)
    assert isinstance(ab.a, A)
    assert isinstance(ab.b, B)


class C:
    pass


class MyABC:
    def __init__(self, ab: AB, c: C):
        self.ab = ab
        self.c = c


def test_container_resolves_recursive_dependencies(container: StaticContainer) -> None:
    abc = container.get(MyABC)
    assert isinstance(abc.ab.a, A)
    assert isinstance(abc.ab.b, B)
    assert isinstance(abc.c, C)


def test_container_resolves_builtins_to_default_values(
    container: StaticContainer
) -> None:
    my_int = container.get(int)
    assert my_int == 0
    my_float = container.get(float)
    assert my_float == 0.0
    my_str = container.get(str)
    assert my_str == ""
    my_bytes = container.get(bytes)
    assert my_bytes == b""
    my_bytes = container.get(bytearray)
    assert my_bytes == bytearray()


def a_func(a: A, b: B, abc: MyABC) -> C:
    return C()


def test_resolve_a_function(container: StaticContainer) -> None:
    my_result = container.get(a_func)
    assert isinstance(my_result, C)


class MyInterface(ABC):
    @abstractmethod
    def foobar(self):
        pass


def test_given_an_abstract_base_type_raises_typeerror(
    container: StaticContainer
) -> None:
    with pytest.raises(TypeError) as e:
        container.get(MyInterface)
    assert "No binding for abstract base" in str(e)
    assert "MyInterface" in str(e)


class MyImplementation(MyInterface):
    def foobar(self):
        pass


def configure_bind(ctx: Config):
    ctx.bind(MyInterface, MyImplementation)


def test_given_a_binding_for_interface_creates_specified_concrete_type() -> None:
    container = create_container(configure_bind)
    concrete_object = container.get(MyInterface)
    assert isinstance(concrete_object, MyImplementation)


class H1(ABC):
    @abstractmethod
    def foobar(self):
        pass


class H2(H1):
    pass


class H3(H2):
    def foobar(self):
        pass


def configure_bind_hierarchy(ctx: Config):
    ctx.bind(H1, H2)
    ctx.bind(H2, H3)


def test_container_resolves_chained_binding() -> None:
    container = create_container(configure_bind_hierarchy)
    concrete_object = container.get(H1)
    assert isinstance(concrete_object, H3)


class MyBaseClass(ABC):
    @abstractmethod
    def foobar(self):
        pass


class NotASubclass:
    pass


def test_bind_a_non_subclass_raises_typeerror() -> None:
    lifetime = Lifetimes(Lifetime.TRANSIENT)
    instances = Instances()
    bindings = Bindings()
    factory_args = FactoryArgs()
    dependencies = Dependencies()
    backend = ConfigBackend(bindings, lifetime, instances, factory_args, dependencies)
    context = Config(backend)
    with pytest.raises(TypeError) as e:
        context.bind(MyBaseClass, NotASubclass)
    assert "{NotASubclass} must be a subclass of {MyBaseClass}".format(
        NotASubclass=NotASubclass, MyBaseClass=MyBaseClass
    ) in str(e)


class MySingleton:
    pass


class MyTransient:
    pass


def configure_lifetime(config: Config):
    config.lifetime(MySingleton, Lifetime.SINGLETON)
    config.lifetime(MyTransient, Lifetime.TRANSIENT)


def test_singleton_lifetimes() -> None:
    container = create_container(configure_lifetime)
    s1 = container.get(MySingleton)
    s2 = container.get(MySingleton)
    assert s1 is s2
    t1 = container.get(MyTransient)
    t2 = container.get(MyTransient)
    assert t1 is not t2


def test_default_lifetime_switched_to_singleton() -> None:
    container = create_container(default_lifetime=Lifetime.SINGLETON)
    s1 = container.get(MySingleton)
    s2 = container.get(MySingleton)
    assert s1 is s2


def test_default_lifetime_is_transient() -> None:
    container = create_container()
    t1 = container.get(MyTransient)
    t2 = container.get(MyTransient)
    assert t1 is not t2


class Transient:
    def __init__(self, a: int, c: C, b: float):
        self.a = a
        self.b = b
        self.c = c


def configure_transient(config: Config):
    config.arguments(Transient, a=42, b=10.0)


def test_transient_creates_transient_object_with_args():
    container = create_container(configure_transient)
    s1 = container.get(Transient)
    assert s1.a == 42
    assert s1.b == 10.0
    assert isinstance(s1.c, C)


def configure_singleton(config: Config):
    config.arguments(Transient, a=42, b=10.0)
    config.lifetime(Transient, Lifetime.SINGLETON)


def test_singleton_creates_singleton_objects_with_args():
    container = create_container(configure_singleton)
    s1 = container.get(Transient)
    s2 = container.get(Transient)
    assert s1.a == 42
    assert s1.b == 10.0
    assert s1 is s2


class MyInstance:
    pass


MY_INSTANCE_INSTANCE = MyInstance()


def configure_instance(config: Config):
    config.instance(MyInstance, MY_INSTANCE_INSTANCE)


def test_instance_returns_provided_intance():
    container = create_container(configure_instance)
    instance = container.get(MyInstance)
    assert instance is MY_INSTANCE_INSTANCE


def get_my_Transient(a: int, c: C, b: float) -> Transient:
    return Transient(a, c, b)


def configure_callable(config: Config):
    config.bind(Transient, get_my_Transient)
    config.arguments(get_my_Transient, a=42)


def test_callable_is_called_with_args_for_return_type_dependency():
    container = create_container(configure_callable)
    instance = container.get(Transient)
    assert instance.a == 42
    assert isinstance(instance.c, C)
    assert instance.b == 0.0


def configure_bind_singleton(config: Config):
    config.bind(MyInterface, MyImplementation)
    config.lifetime(MyImplementation, Lifetime.SINGLETON)


def test_lifetime_of_binded_type_is_singleton():
    container = create_container(configure_bind_singleton)
    s1 = container.get(MyInterface)
    s2 = container.get(MyInterface)
    assert s1 is s2


class Dependecy:
    pass


class NeedsDependency:
    def __init__(self, dependet: Dependecy):
        self.dependent = dependet


def configure_dependency(config: Config):
    config.dependency(Dependecy)


def test_given_dependencies_are_used_for_resolution():
    dependency = Dependecy()
    container = create_container(configure_dependency, dependencies=[dependency])
    dependent = container.get(NeedsDependency)
    assert dependent.dependent is dependency


class F(ABC):
    @abstractmethod
    def someting(self):
        pass


class F1(F):
    def someting(self):
        pass


class F2(F):
    def someting(self):
        pass


class UseF1:
    def __init__(self, f: F):
        self.f = f


class UseF2:
    def __init__(self, f: F):
        self.f = f


def configure_binding_with_context(config: Config):
    config.bind(F, F1, where=UseF1)
    config.bind(F, F2, where=UseF2)


def test_binding_within_a_context():
    container = create_container(configure_binding_with_context)
    f1 = container.get(UseF1)
    f2 = container.get(UseF2)
    assert isinstance(f1.f, F1)
    assert isinstance(f2.f, F2)


class T:
    def __init__(self, a: int):
        self.a = a


class UseT1:
    def __init__(self, t: T):
        self.t = t


class UseT2:
    def __init__(self, t: T):
        self.t = t


def configure_configure_transient_with_context(config: Config):
    config.arguments(T, a=1, where=UseT1)
    config.arguments(T, a=2, where=UseT2)


def test_transient_within_a_context():
    container = create_container(configure_configure_transient_with_context)
    t1 = container.get(UseT1)
    t2 = container.get(UseT2)
    assert t1.t.a == 1
    assert t2.t.a == 2


def configure_singleton_with_context(config: Config):
    config.arguments(T, a=1, where=UseT1)
    config.arguments(T, a=2, where=UseT2)
    config.lifetime(T, Lifetime.SINGLETON, where=UseT1)
    config.lifetime(T, Lifetime.SINGLETON, where=UseT2)


def test_singleton_within_a_context():
    container = create_container(configure_singleton_with_context)
    t11 = container.get(UseT1)
    t12 = container.get(UseT1)
    t21 = container.get(UseT2)
    t22 = container.get(UseT2)
    assert t11.t.a == 1
    assert t11.t is t12.t
    assert t21.t.a == 2
    assert t21.t is t22.t


t1_instance = T(1)
t2_instance = T(2)


def configure_instance_with_context(config: Config):
    config.instance(T, t1_instance, where=UseT1)
    config.instance(T, t2_instance, where=UseT2)


def test_instance_within_a_context():
    container = create_container(configure_instance_with_context)
    t1 = container.get(UseT1)
    t2 = container.get(UseT2)
    assert t1.t is t1_instance
    assert t2.t is t2_instance


def foo(a: int) -> T:
    return T(a)


def configure_callable_with_context(config: Config):
    config.bind(T, foo)
    config.arguments(foo, a=1, where=UseT1)
    config.arguments(foo, a=2, where=UseT2)


def test_callable_within_a_context():
    container = create_container(configure_callable_with_context)
    t1 = container.get(UseT1)
    t2 = container.get(UseT2)
    assert t1.t.a == 1
    assert t2.t.a == 2


class UseT3:
    def __init__(self, t: T):
        self.t = t


def configure_lifetime_with_context(config: Config):
    config.bind(T, foo)
    config.arguments(foo, a=1, where=UseT1)
    config.arguments(foo, a=2, where=UseT2)
    config.lifetime(foo, Lifetime.SINGLETON, where=UseT2)
    config.lifetime(foo, where=UseT3, lifetime=Lifetime.SINGLETON)


def test_scope_setting_within_a_context():
    container = create_container(configure_lifetime_with_context)
    t11 = container.get(UseT1)
    t12 = container.get(UseT1)
    t21 = container.get(UseT2)
    t22 = container.get(UseT2)
    assert t11.t.a == 1
    assert t11.t is not t12.t
    assert t21.t.a == 2
    assert t21.t is t22.t

    t31 = container.get(UseT3)
    t32 = container.get(UseT3)
    assert t31.t is t32.t


def test_resolver_raises_assertion_error_if_no_handler_is_found():
    resolver = Resolver()
    with pytest.raises(AssertionError):
        resolver.get_instance(int)


def configure_without_dependencies(config: Config):
    pass


def test_give_a_non_registered_dependency_raises_type_error():
    with pytest.raises(TypeError):
        create_container(configure_without_dependencies, dependencies=[5])


def test_define_a_dependency_without_providing_it_raises_typeerror():
    def configure(config: Config):
        config.dependency(int)

    with pytest.raises(TypeError):
        create_container(configure)


def test_callable_which_returns_none_raises_typeerror():
    def foo():
        pass

    def configure(config: Config):
        config.bind(T, foo)

    with pytest.raises(TypeError):
        create_container(configure)


class SB1:
    pass


class SB2(SB1):
    pass


class SB3(SB2):
    pass


class MySB:
    def __init__(self, sb: SB1):
        self.sb = sb


def configure_stacked_binding_and_lifetime(config: Config):
    config.bind(SB1, SB2)
    config.lifetime(SB2, Lifetime.SINGLETON)
    config.bind(SB2, SB3)


def test_stacked_binding_uses_singleton_of_bound_to_object_first():
    container = create_container(configure_stacked_binding_and_lifetime)
    s1 = container.get(MySB)
    s2 = container.get(MySB)
    assert isinstance(s1.sb, SB3)
    assert s1.sb is not s2.sb


class MyArg:
    def __init__(self, arg1):
        self.arg1 = arg1


def configure_arguments(config: Config):
    config.arguments(MyArg, arg1="foobar")


def test_provide_arguments_for_a_callable_container():
    container = create_container(configure_arguments)
    my_arg = container.get(MyArg)
    assert my_arg.arg1 == "foobar"


class ArgFactory:
    def __init__(self, name: str):
        self.name = name


class NameProvider:
    def __init__(self, name: str):
        self.name = name


def factory(provider: NameProvider) -> str:
    return provider.name


def configure_arg_factory(config: Config):
    config.arg_factory(ArgFactory, name=factory)
    config.arguments(NameProvider, name="foobar")


def test_arg_factory():
    container = create_container(configure_arg_factory)
    my_arg = container.get(ArgFactory)
    assert my_arg.name == "foobar"


class ParameterAndFactory:
    def __init__(self, number: int, text: str):
        self.number = number
        self.text = text


def get_text() -> str:
    return "foobar"


def configure_parameter_and_factory(config: Config):
    config.arguments(ParameterAndFactory, number=42, text="42")
    config.arg_factory(ParameterAndFactory, text=get_text)


def test_parameter_and_factory():
    container = create_container(configure_parameter_and_factory)
    my_arg = container.get(ParameterAndFactory)
    assert my_arg.number == 42
    assert my_arg.text == "foobar"


class MyInt:
    def get_int(self) -> int:
        return 42


class ProvidesInt:
    def __init__(self, a_int: MyInt):
        self._a_int = a_int

    def get_int(self) -> int:
        return self._a_int.get_int()

    @staticmethod
    def get_int_static() -> int:
        return 10


class NeedsInt:
    def __init__(self, a_int: int):
        self.a_int = a_int


def configure_factory_class_static(config: Config):
    config.arg_factory(NeedsInt, a_int=ProvidesInt.get_int_static)


def test_container_factory_is_class_static_method():
    container = create_container(configure_factory_class_static)
    needs_int = container.get(NeedsInt)
    assert needs_int.a_int == 10


def configure_factory_class_instance_method(config: Config):
    config.arg_factory(NeedsInt, a_int=ProvidesInt(MyInt()).get_int)


def test_container_factory_is_class_instance_method():
    container = create_container(configure_factory_class_instance_method)
    needs_int = container.get(NeedsInt)
    assert needs_int.a_int == 42


def configure_factory_not_created_class_instance_method(config: Config):
    config.arg_factory(NeedsInt, a_int=ProvidesInt.get_int)


def test_container_factory_is_not_created_class_instance_method():
    container = create_container(configure_factory_not_created_class_instance_method)
    needs_int = container.get(NeedsInt)
    assert needs_int.a_int == 42


class MethodClass:
    def bar(self):
        pass

    @staticmethod
    def foo():
        pass

    @classmethod
    def foobar(cls) -> "MethodClass":
        return MethodClass()

    def __call__(self):
        pass


def some_function():
    pass


def test_method_is_of_a_not_created_class():
    assert method_of_not_created_class(MethodClass.bar) is True
    assert method_of_not_created_class(MethodClass.foo) is False
    assert method_of_not_created_class(MethodClass.foobar) is False
    assert method_of_not_created_class(MethodClass().bar) is False
    assert method_of_not_created_class(MethodClass().foo) is False
    assert method_of_not_created_class(MethodClass().foobar) is False
    assert method_of_not_created_class(some_function) is False
    assert method_of_not_created_class(MethodClass()) is False



__version__ = "__version__ = '0.0.6'"

from smart_injector.config.user import Config
from smart_injector.container.container import StaticContainer
from smart_injector.container.factory import create_container
from smart_injector.lifetime import Lifetime

__all__ = ["create_container", "StaticContainer", "Lifetime", "Config"]



from typing import Callable
from typing import List
from typing import TypeVar
from typing import cast

from smart_injector.types import Handler
from smart_injector.types import ResolveRequest

T = TypeVar("T")
S = TypeVar("S")


class Resolver:
    def __init__(self):
        self._type_handlers = []  # type: List[Handler]

    def add_type_handler(self, handler: Handler):
        self._type_handlers.append(handler)

    def get_instance(self, a_type: Callable[..., T]) -> T:
        return self.get_new_instance(ResolveRequest(a_type, a_type, None))

    def get_new_instance(self, context: ResolveRequest) -> T:
        for handler in cast(
            List[Handler], self._type_handlers
        ):  # use list cast to surpress pylama List not used warning
            if handler.can_handle_type(context):
                return handler.handle(context)
        assert False, "should not reach this. you should have added a default handler"



from enum import Enum


class Lifetime(Enum):
    """Specifies the lifetime for objects created by the container

    :Lifetime.SINGLETON: :py:meth:`smart_injector.StaticContainer.get` returns the same every instance on every call
    :Lifetime.TRANSIENT: :py:meth:`smart_injector.StaticContainer.get` returns a new instance on every call
    """

    SINGLETON = 0
    TRANSIENT = 1
    _INTERNAL_DEFAULT = 2



import inspect
from typing import Callable
from typing import Optional
from typing import Type
from typing import TypeVar

T = TypeVar("T")


def get_return_type(a_type: Callable[..., T]) -> Optional[Type[T]]:
    r_type = inspect.signature(a_type).return_annotation
    if r_type is inspect.Signature.empty:
        return None
    return r_type
